use crate::write_log;

use self::{
    rendering::Rendering,
    rfactor_data::{DataReceiver, RFactorData},
};

mod elements;
mod rendering;
mod rfactor_data;

use anyhow::{bail, Result};
use assetpath::AssetPath;
use std::{
    collections::HashMap,
    sync::{Arc, Mutex},
};
use ui::{guihandler::guihandler::Font, prelude::*};
use vulkan_rs::prelude::*;

use elements::*;

use serde::{Deserialize, Serialize};

pub trait UiOverlay: DataReceiver {}

#[derive(Deserialize, Serialize, Debug)]
pub struct UiSelectorConfig {
    pub enable_watermark: bool,

    pub enable_radar: bool,
    pub enable_pedals: bool,
    pub enable_leaderboard: bool,
}

impl UiSelectorConfig {
    pub const fn new() -> Self {
        Self {
            enable_watermark: true,
            enable_radar: true,
            enable_pedals: true,
            enable_leaderboard: true,
        }
    }
}

#[derive(Deserialize, Serialize, Debug)]
pub struct OverlayConfig {
    pub ui_config: UiSelectorConfig,
    pub radar_config: RadarConfig,
    pub leader_board_config: LeaderBoardConfig,
}

impl OverlayConfig {
    pub const fn new() -> Self {
        Self {
            ui_config: UiSelectorConfig::new(),
            radar_config: RadarConfig::new(),
            leader_board_config: LeaderBoardConfig::new(),
        }
    }
}

#[derive(Debug)]
pub struct InstanceCollection {
    pub instance: Arc<Instance>,
    pub physical_devices: Vec<VkPhysicalDevice>,
    pub device_collection: HashMap<VkDevice, DeviceCollection>,
}

#[derive(Debug)]
pub struct DeviceCollection {
    pub device: Arc<Device>,
    pub queues: HashMap<VkQueue, Arc<Mutex<Queue>>>,
    pub swapchains: HashMap<VkSwapchainKHR, Arc<Swapchain>>,
}

pub struct Overlay {
    config: OverlayConfig,

    pub(crate) instances: HashMap<VkInstance, InstanceCollection>,
    rendering: Option<Rendering>,
    gui_handler: Option<Arc<GuiHandler>>,

    ui_elements: Vec<Arc<Mutex<dyn UiOverlay>>>,

    rfactor_data: Option<RFactorData>,
}

impl Overlay {
    pub fn new() -> Self {
        Self {
            config: OverlayConfig::new(),

            instances: HashMap::new(),
            rendering: None,
            gui_handler: None,
            ui_elements: Vec::new(),

            rfactor_data: None,
        }
    }

    pub fn set_config(&mut self, config: OverlayConfig) {
        self.config = config;
    }

    pub fn add_instance(&mut self, instance: Arc<Instance>) {
        let physical_devices = match instance.enumerate_physical_devices() {
            Ok(devices) => devices,
            Err(err) => {
                write_log!(format!("failed to enumerate physical devices {err:?}"));
                panic!();
            }
        };

        self.instances.insert(
            instance.vk_handle(),
            InstanceCollection {
                instance,
                physical_devices,
                device_collection: HashMap::new(),
            },
        );
    }

    pub fn remove_instance(&mut self, handle: VkInstance) {
        self.instances.remove(&handle);
    }

    pub fn instance_by_physical_device(&self, physical_device: VkPhysicalDevice) -> Arc<Instance> {
        match self
            .instances
            .values()
            .find(|collection| {
                collection
                    .physical_devices
                    .iter()
                    .find(|&&phys_dev| phys_dev == physical_device)
                    .is_some()
            })
            .map(|collection| &collection.instance)
        {
            Some(instance) => instance.clone(),
            None => {
                write_log!(format!(
                    "failed to find instance with physical device ({physical_device:?})"
                ));
                panic!();
            }
        }
    }

    pub fn set_device(&mut self, device: Arc<Device>) {
        if let Some(instance_collection) = self
            .instances
            .get_mut(&device.physical_device().instance().vk_handle())
        {
            instance_collection.device_collection.insert(
                device.vk_handle(),
                DeviceCollection {
                    device,
                    queues: HashMap::new(),
                    swapchains: HashMap::new(),
                },
            );
        }
    }

    fn device_collection(&self, device: VkDevice) -> &DeviceCollection {
        for instance_collection in self.instances.values() {
            if let Some(device_collection) = instance_collection.device_collection.get(&device) {
                return device_collection;
            }
        }

        write_log!(format!("device ({device:?}) not found"));
        panic!();
    }

    pub fn device_collection_mut(&mut self, device: &Arc<Device>) -> &mut DeviceCollection {
        if let Some(instance_collection) = self
            .instances
            .get_mut(&device.physical_device().instance().vk_handle())
        {
            if let Some(device_collection) = instance_collection
                .device_collection
                .get_mut(&device.vk_handle())
            {
                return device_collection;
            }
        }

        write_log!(format!("device ({device:?}) not found"));
        panic!();
    }

    pub fn device(&self, handle: VkDevice) -> Arc<Device> {
        self.device_collection(handle).device.clone()
    }

    pub fn set_queue(&mut self, queue: Arc<Mutex<Queue>>) {
        let handle = queue.lock().unwrap().vk_handle();
        let device = queue.lock().unwrap().device().clone();

        self.device_collection_mut(&device)
            .queues
            .insert(handle, queue);
    }

    pub fn insert_swapchain(&mut self, swapchain: Arc<Swapchain>) {
        self.device_collection_mut(swapchain.device())
            .swapchains
            .insert(swapchain.vk_handle(), swapchain);
    }

    pub fn swapchain(&self, swapchain: VkSwapchainKHR) -> Option<&Arc<Swapchain>> {
        let sc = self.rendering.as_ref().unwrap().swapchain();

        if sc.vk_handle() == swapchain {
            Some(sc)
        } else {
            None
        }
    }

    pub fn create_rendering(&mut self, device: VkDevice) -> Result<()> {
        write_log!("-> create rendering: start");

        self.rendering = None;

        write_log!("-> create rendering: old cleared");

        let device_collection = self.device_collection(device);

        let device = &device_collection.device.clone();
        let queue = match device_collection.queues.values().next() {
            Some(queue) => queue.clone(),
            None => {
                write_log!(format!("failed to find queue"));
                panic!();
            }
        };
        let swapchain = match device_collection.swapchains.values().next() {
            Some(swapchain) => swapchain.clone(),
            None => {
                write_log!(format!("failed to find swapchain"));
                panic!();
            }
        };

        let mut rendering = Rendering::new(queue.clone(), swapchain.clone())?;

        write_log!("-> create rendering: new created");

        // only font is used
        let mut create_info = GuiHandlerCreateInfo::default();

        create_info.font = Font::Bytes(include_bytes!("../../resources/font.png"));

        // required to not crash
        create_info.resource_directory = AssetPath::from("");
        create_info.resource_directory.assume_prefix_free();

        // provide trait required by GuiHandler
        let ctx = Arc::new(ContextImpl::new(
            device.clone(),
            queue.clone(),
            swapchain.clone(),
            rendering.images().clone(),
        ));

        write_log!("create rendering: context created");

        // create GuiHandler
        let gui_handler = GuiHandler::new(create_info, &(ctx as Arc<dyn ContextInterface>))?;
        write_log!("create rendering: GuiHandler successfully created");

        // create ui elements

        // create watermark
        if self.config.ui_config.enable_watermark {
            let watermark = Arc::new(Mutex::new(Watermark::new(&gui_handler)?));

            self.ui_elements.push(watermark);

            write_log!("create rendering: Watermark successfully created");
        }

        // create radar
        if self.config.ui_config.enable_radar {
            let radar = Arc::new(Mutex::new(Radar::new(
                self.config.radar_config,
                device.clone(),
                queue.clone(),
                &rendering,
            )?));

            rendering.add_render_callback({
                let radar = radar.clone();

                move |index| radar.lock().unwrap().render(index)
            });

            self.ui_elements.push(radar);

            write_log!("create rendering: Radar successfully created");
        }

        // create pedals
        if self.config.ui_config.enable_pedals {
            let pedals = Arc::new(Mutex::new(Pedals::new(
                &gui_handler,
                device.clone(),
                queue.clone(),
            )?));

            self.ui_elements.push(pedals.clone());

            rendering.add_render_callback(move |_| pedals.lock().unwrap().render());

            write_log!("create rendering: Pedals successfully created");
        }

        // create leaderboard
        if self.config.ui_config.enable_leaderboard {
            let leaderboard = Arc::new(Mutex::new(LeaderBoard::new(
                &gui_handler,
                self.config.leader_board_config,
            )?));
            self.ui_elements.push(leaderboard);

            write_log!("create rendering: Leader Board successfully created");
        }

        // add rendering callbacks
        rendering.add_render_callback({
            let gui_handler = gui_handler.clone();
            let device = device.clone();
            let queue = queue.clone();

            move |index| {
                let command_buffer =
                    CommandBuffer::new_primary().build(device.clone(), queue.clone())?;

                {
                    let mut recorder = command_buffer.begin(VkCommandBufferBeginInfo::new(
                        VK_COMMAND_BUFFER_USAGE_ONE_TIME_SUBMIT_BIT
                            | VK_COMMAND_BUFFER_USAGE_SIMULTANEOUS_USE_BIT,
                    ))?;

                    gui_handler.process(&mut recorder, &TargetMode::Mono(index as usize))?;
                }

                Ok(command_buffer)
            }
        });

        write_log!("create rendering: render callbacks added");

        self.rendering = Some(rendering);
        self.gui_handler = Some(gui_handler);

        write_log!("-> create rendering: end");

        Ok(())
    }

    pub fn render(
        &mut self,
        wait_semaphores: &[VkSemaphore],
        swapchain: &Arc<Swapchain>,
    ) -> Result<&Arc<Semaphore>> {
        if self.rfactor_data.is_none() {
            self.rfactor_data = RFactorData::new().ok();

            if let Some(data) = &mut self.rfactor_data {
                write_log!("created RFactorData");

                for receiver in self.ui_elements.iter() {
                    data.add_receiver(receiver.clone());
                }
            }
        }

        if let Some(rfactor) = &mut self.rfactor_data {
            rfactor.update()?;
        }

        let rendering = self.rendering.as_ref().unwrap();

        if swapchain.vk_handle() != rendering.swapchain().vk_handle() {
            bail!("swapchains did not match");
        } else {
            rendering.render(wait_semaphores)
        }
    }
}

struct ContextImpl {
    device: Arc<Device>,
    queue: Arc<Mutex<Queue>>,
    swapchain: Arc<Swapchain>,
    images: Vec<Arc<Image>>,
}

impl ContextImpl {
    fn new(
        device: Arc<Device>,
        queue: Arc<Mutex<Queue>>,
        swapchain: Arc<Swapchain>,
        images: Vec<Arc<Image>>,
    ) -> Self {
        Self {
            device,
            queue,
            swapchain,
            images,
        }
    }
}

impl ContextInterface for ContextImpl {
    fn device(&self) -> &Arc<Device> {
        &self.device
    }

    fn queue(&self) -> &Arc<Mutex<Queue>> {
        &self.queue
    }

    fn format(&self) -> VkFormat {
        self.swapchain.format()
    }

    fn image_layout(&self) -> VkImageLayout {
        VK_IMAGE_LAYOUT_PRESENT_SRC_KHR
    }

    fn image_count(&self) -> usize {
        self.images.len()
    }

    fn images(&self) -> TargetMode<Vec<Arc<Image>>> {
        TargetMode::Mono(self.images.clone())
    }

    fn width(&self) -> u32 {
        self.swapchain.width()
    }

    fn height(&self) -> u32 {
        self.swapchain.height()
    }
}
