mod enums;
mod structs;
mod vk_handles;

use crate::check_logging;
use enums::*;
use structs::*;
use vk_handles::*;
use vulkan_rs::prelude::*;

use std::{mem, os::raw::c_char, ptr};

use crate::{get_config, logging, overlay, write_log};

#[no_mangle]
#[allow(non_snake_case)]
pub(crate) extern "C" fn vkNegotiateLoaderLayerInterfaceVersion(
    pVersionStruct: *mut VkNegotiateLayerInterface,
) -> VkResult {
    let home = std::env::var("HOME").unwrap();

    check_logging(&home);

    overlay().set_config(get_config(&home));

    unsafe {
        *pVersionStruct = VkNegotiateLayerInterface {
            sType: enums::VkNegotiateLayerStructType::LAYER_NEGOTIATE_INTERFACE_STRUCT,
            pNext: ptr::null_mut(),
            loaderLayerInterfaceVersion: 2,
            pfnGetInstanceProcAddr: Some(get_instance_proc_addr),
            pfnGetDeviceProcAddr: Some(get_device_proc_addr),
            pfnGetPhysicalDeviceProcAddr: None,
        }
    };

    set_vk_handles(match VkTypedefHandles::new() {
        Ok(handles) => handles,
        Err(err) => {
            write_log!(format!("failed to load typedef handles {:?}", err));
            return VK_ERROR_INITIALIZATION_FAILED;
        }
    });

    VK_SUCCESS
}

#[no_mangle]
extern "system" fn get_device_proc_addr(
    _device: VkDevice,
    function_name: *const c_char,
) -> PFN_vkVoidFunction {
    if cfg!(debug_assertions) {
        write_log!("get_device_proc_addr");
    }

    let func_string = match VkString::try_from(function_name) {
        Ok(func) => func,
        Err(_) => {
            write_log!("Err: failed creating string");
            return Functions::Null.convert();
        }
    };

    let s = func_string.as_str();

    if cfg!(debug_assertions) {
        write_log!(format!("\t{}", s));
    }

    if let Some(func) = Functions::get_vk_func(s) {
        return func.convert();
    }

    if let Some(func) = vk_handles().handle(s) {
        return func;
    }

    if cfg!(debug_assertions) {
        write_log!("\t-> returning null");
    }

    Functions::Null.convert()
}

#[no_mangle]
extern "system" fn get_instance_proc_addr(
    _instance: VkInstance,
    function_name: *const c_char,
) -> PFN_vkVoidFunction {
    if cfg!(debug_assertions) {
        write_log!("get_instance_proc_addr");
    }

    let func_string = match VkString::try_from(function_name) {
        Ok(func) => func,
        Err(_) => {
            write_log!("Err: failed creating string");
            return Functions::Null.convert();
        }
    };

    let s = func_string.as_str();

    if cfg!(debug_assertions) {
        write_log!(format!("\t{}", s));
    }

    if let Some(func) = Functions::get_vk_func(s) {
        return func.convert();
    }

    if let Some(func) = vk_handles().handle(s) {
        return func;
    }

    if cfg!(debug_assertions) {
        write_log!("\t-> returning null");
    }

    Functions::Null.convert()
}

pub(crate) extern "system" fn create_instance(
    create_info: *const VkInstanceCreateInfo,
    allocator: *const VkAllocationCallbacks,
    instance: *mut VkInstance,
) -> VkResult {
    write_log!(" ================== vulkan layer create instance ==================");

    let chain_info = match VkLayerInstanceCreateInfo::get_chain_info(
        unsafe { &*create_info },
        VK_LAYER_LINK_INFO,
    ) {
        Some(info) => info,
        None => {
            write_log!("instance chain info not found.");
            return VK_ERROR_LAYER_NOT_PRESENT;
        }
    };

    let proc_addr = chain_info.layer_info().next_instance_proc_addr;
    let create_instance: PFN_vkCreateInstance = unsafe {
        mem::transmute(proc_addr(
            VkInstance::NULL_HANDLE,
            VkString::new("vkCreateInstance").as_ptr(),
        ))
    };

    chain_info.advance_layer_info();
    let result = create_instance(create_info, allocator, instance);
    if result != VK_SUCCESS {
        return result;
    };

    vk_handles_mut().load_instance_functions(unsafe { *instance }, proc_addr);

    write_log!("-> successfully created instance.");

    let ext_names = unsafe { (*create_info).extension_names() };

    write_log!(format!("{:?}", ext_names));

    // DXVK workaround, it creates the instance twice with different properties
    if ext_names.contains(&VkString::new("VK_KHR_surface")) {
        write_log!("found VK_KHR_surface");

        unsafe {
            let panic_result =
                match std::panic::catch_unwind(|| -> anyhow::Result<std::sync::Arc<Instance>> {
                    Instance::preinitialized(
                        *instance,
                        proc_addr,
                        &ext_names,
                        if (*create_info).pApplicationInfo == ptr::null() {
                            None
                        } else {
                            Some((*(*create_info).pApplicationInfo).apiVersion)
                        },
                    )
                }) {
                    Ok(panic_ok) => panic_ok,
                    Err(err) => {
                        write_log!(format!(
                            "catched a panic from Instance::preinitialized: {:?}",
                            err
                        ));
                        return VK_ERROR_INITIALIZATION_FAILED;
                    }
                };

            let ins = match panic_result {
                Ok(ins) => ins,
                Err(err) => {
                    write_log!(format!("-> local instance creation failed: {:?}", err));
                    return VK_ERROR_INITIALIZATION_FAILED;
                }
            };

            overlay().add_instance(ins);
        }

        write_log!("-> created local instance handle");
    }

    write_log!("returning from create_instance");

    VK_SUCCESS
}

pub(crate) extern "system" fn destroy_instance(
    instance: VkInstance,
    allocator: *const VkAllocationCallbacks,
) {
    write_log!(" ================== vulkan layer destroy instance ==================");

    unsafe {
        if let Some(vk_fn) = vk_handles().handle("vkDestroyInstance") {
            let destroy_instance: PFN_vkDestroyInstance = mem::transmute(vk_fn);
            overlay().remove_instance(instance);

            destroy_instance(instance, allocator);
        }
    }
}

pub(crate) extern "system" fn create_device(
    physical_device: VkPhysicalDevice,
    create_info: *const VkDeviceCreateInfo<'_>,
    allocator: *const VkAllocationCallbacks,
    device: *mut VkDevice,
) -> VkResult {
    write_log!(" ================== vulkan layer create device ==================");

    let chain_info =
        match VkLayerDeviceCreateInfo::get_chain_info(unsafe { &*create_info }, VK_LAYER_LINK_INFO)
        {
            Some(info) => info,
            None => {
                write_log!("device chain info not found.");
                return VK_ERROR_LAYER_NOT_PRESENT;
            }
        };

    let proc_addr = chain_info.layer_info().next_device_proc_addr;

    chain_info.advance_layer_info();

    let result = unsafe {
        match vk_handles().handle("vkCreateDevice") {
            Some(vk_fn) => {
                let create_device: PFN_vkCreateDevice = mem::transmute(vk_fn);
                create_device(physical_device, create_info, allocator, device)
            }
            None => {
                write_log!("failed creating device");
                return VK_ERROR_INITIALIZATION_FAILED;
            }
        }
    };

    if result != VK_SUCCESS {
        return result;
    }

    vk_handles_mut().load_device_functions(unsafe { *device }, proc_addr);

    let mut overlay = overlay();

    let instance = overlay.instance_by_physical_device(physical_device);

    let pdev = match PhysicalDevice::from_raw(instance, physical_device) {
        Ok(pdev) => pdev,
        Err(err) => {
            write_log!(format!("failed creating physical device: {:?}", err));
            return VK_ERROR_INITIALIZATION_FAILED;
        }
    };

    let queue_info = match Queue::create_non_presentable_request_info(&pdev, VK_QUEUE_GRAPHICS_BIT)
    {
        Ok(qi) => qi,
        Err(err) => {
            write_log!(format!("failed getting queue info: {:?}", err));
            return VK_ERROR_INITIALIZATION_FAILED;
        }
    };

    unsafe {
        let create_queue_info = std::slice::from_raw_parts(
            (*create_info).pQueueCreateInfos,
            (*create_info).queueCreateInfoCount as usize,
        );

        for info in create_queue_info {
            write_log!(format!(
                "pCreateDeviceInfo; queue fam: {}, queue count: {}",
                info.queueFamilyIndex, info.queueCount
            ));
        }

        write_log!(format!(
            "Queue: queue fam: {}, queue count: {}",
            queue_info.queue_create_info.queueFamilyIndex, queue_info.queue_create_info.queueCount
        ));
    }

    let ext_names = unsafe { (*create_info).extension_names() };

    write_log!(format!("{:?}", ext_names));

    let device = match Device::preinitialized(unsafe { *device }, proc_addr, pdev, &ext_names) {
        Ok(dev) => dev,
        Err(err) => {
            write_log!(format!("-> local device creation failed: {:?}", err));
            return VK_ERROR_INITIALIZATION_FAILED;
        }
    };

    write_log!("created device");

    let queue = device.get_queue(queue_info.queue_family_index, queue_info.queue_index);

    write_log!("got queue from device");

    overlay.set_device(device.clone());
    overlay.set_queue(queue);

    VK_SUCCESS
}

pub(crate) extern "system" fn destroy_device(
    device: VkDevice,
    allocator: *const VkAllocationCallbacks,
) {
    write_log!(" ================== vulkan layer destroy device ==================");

    unsafe {
        if let Some(vk_fn) = vk_handles().handle("vkDestroyDevice") {
            let destroy_device: PFN_vkDestroyDevice = mem::transmute(vk_fn);

            destroy_device(device, allocator);
        }
    }
}

pub(crate) extern "system" fn create_swapchain(
    _device: VkDevice,
    create_info: *const VkSwapchainCreateInfoKHR,
    _allocator: *const VkAllocationCallbacks,
    p_swapchain: *mut VkSwapchainKHR,
) -> VkResult {
    write_log!(" ================== vulkan layer create swapchain ==================");

    let create_swapchain: PFN_vkCreateSwapchainKHR =
        match vk_handles().handle("vkCreateSwapchainKHR") {
            Some(create_swapchain) => unsafe { mem::transmute(create_swapchain) },
            None => {
                write_log!("failed querying vkCreateSwapchainKHR");
                return VK_ERROR_INITIALIZATION_FAILED;
            }
        };

    create_swapchain(_device, create_info, _allocator, p_swapchain);

    write_log!(format!("-> created swapchain vk handle {:?}", unsafe {
        *p_swapchain
    }));

    let mut overlay = overlay();

    let swapchain =
        unsafe { Swapchain::from_raw(overlay.device(_device), &*create_info, *p_swapchain) };
    overlay.insert_swapchain(swapchain);

    {
        let instances = &overlay.instances;

        write_log!(format!("number of instances: {}", instances.len()));

        for instance_collection in instances.values() {
            write_log!(format!(
                "\tnumber of physical devices: {}",
                instance_collection.physical_devices.len()
            ));

            write_log!(format!(
                "\tnumber of devices: {}",
                instance_collection.device_collection.len()
            ));

            for device_collection in instance_collection.device_collection.values() {
                write_log!(format!(
                    "\t\tnumber of queues: {}",
                    device_collection.queues.len()
                ));
                write_log!(format!(
                    "\t\tnumber of swapchains: {}",
                    device_collection.swapchains.len()
                ));
            }
        }
    }

    write_log!("-> created Arc<Swapchain>");

    if let Err(err) = overlay.create_rendering(_device) {
        write_log!(format!("create overlay rendering struct failed: {:?}", err));
        return VK_ERROR_INITIALIZATION_FAILED;
    }

    write_log!("-> created renderer");

    VK_SUCCESS
}

pub(crate) extern "system" fn destroy_swapchain(
    device: VkDevice,
    swapchain: VkSwapchainKHR,
    allocator: *const VkAllocationCallbacks,
) -> () {
    write_log!(" ================== vulkan layer destroy swapchain ==================");

    unsafe {
        if let Some(vk_fn) = vk_handles().handle("vkDestroySwapchainKHR") {
            let destroy_swapchain: PFN_vkDestroySwapchainKHR = mem::transmute(vk_fn);

            destroy_swapchain(device, swapchain, allocator);
        }

        let mut overlay = overlay();
        let device = overlay.device(device);

        overlay
            .device_collection_mut(&device)
            .swapchains
            .remove(&swapchain);
    }
}

pub(crate) extern "system" fn present_queue(
    queue: VkQueue,
    present_info: *const VkPresentInfoKHR,
) -> VkResult {
    if logging() {
        write_log!(" ================== vulkan layer queue present ==================");
        unsafe {
            let swapchains = std::slice::from_raw_parts(
                (*present_info).pSwapchains,
                (*present_info).swapchainCount as usize,
            );

            write_log!(format!("present {} swapchain(s)", swapchains.len()));

            for swapchain in swapchains {
                write_log!(format!("present swapchain: {:?}", swapchain));

                if let Some(swch) = overlay().swapchain(*swapchain) {
                    write_log!(format!(
                        " -> internal swapchain found! ({:?})",
                        swch.vk_handle()
                    ));
                }
            }
        }
    }

    let mut overlay = overlay();
    let wait_semaphores = unsafe { (*present_info).wait_semaphores() };

    let pfn: PFN_vkQueuePresentKHR = match vk_handles().handle("vkQueuePresentKHR") {
        Some(pfn) => unsafe { mem::transmute(pfn) },
        None => {
            write_log!("failed querying vkQueuePresentKHR");
            return VK_ERROR_DEVICE_LOST;
        }
    };

    for (index, swapchain_info) in unsafe { (*present_info).swapchain_info() }
        .iter()
        .enumerate()
    {
        if let Some(swapchain) = overlay.swapchain(swapchain_info.swapchain).cloned() {
            unsafe { swapchain.set_image_index(swapchain_info.index) };
            let swapchain_handle = swapchain.vk_handle();

            match overlay.render(wait_semaphores, &swapchain) {
                Ok(signal_semaphore) => {
                    let wait_semaphores = &[signal_semaphore.vk_handle()];
                    let swapchain_handles = &[swapchain_handle];
                    let indices = &[swapchain_info.index];

                    let mut new_present_info =
                        VkPresentInfoKHR::new(wait_semaphores, swapchain_handles, indices, &mut []);
                    new_present_info.pNext = unsafe { (*present_info).pNext };

                    if signal_semaphore.vk_handle() != new_present_info.wait_semaphores()[0] {
                        write_log!(format!(
                            "wait_semaphore: expected ({:?}) - found ({:?})",
                            signal_semaphore.vk_handle(),
                            new_present_info.wait_semaphores()[0],
                        ));
                    }

                    if swapchain_handle != new_present_info.swapchain_info()[0].swapchain {
                        write_log!(format!(
                            "swapchain: expected ({:?}) - found ({:?})",
                            swapchain_handle,
                            new_present_info.swapchain_info()[0].swapchain,
                        ));
                    }

                    if swapchain_info.index != new_present_info.swapchain_info()[0].index {
                        write_log!(format!(
                            "index: expected ({:?}) - found ({:?})",
                            swapchain_info.index,
                            new_present_info.swapchain_info()[0].index,
                        ));
                    }

                    write_log!(format!("present_info: {new_present_info:#?}"));
                    unsafe {
                        write_log!(format!("present_info: {:#?}", *present_info));
                    }

                    let res = pfn(queue, &new_present_info);

                    let results = unsafe { (*present_info).results() };

                    if results.len() > index {
                        results[index] = res;
                    }

                    if res != VK_SUCCESS {
                        write_log!(format!("queue present failed: {:?}", res));
                    }
                }
                Err(err) => {
                    write_log!(format!("overlay rendering failed: {:?}", err));
                    return VK_ERROR_DEVICE_LOST;
                }
            }
        }
    }

    write_log!("finish present queue");

    match unsafe { (*present_info).results() }
        .iter()
        .find(|&&result| result != VK_SUCCESS)
    {
        Some(&error) => error,
        None => VK_SUCCESS,
    }
}
